import {
  MONITORING_SYSTEM_API_VERSION, KIBANA_SYSTEM_ID, KIBANA_STATS_TYPE
} from '../../../common/constants';
import { rollupEvent } from './map_event';
import { sourceKibana } from './source_kibana';
import { monitoringBulk } from './monitoring_bulk';
import { CloudDetector } from '../../cloud';

/**
 * Manage the buffer of Kibana Ops events
 * Does the bulk upload to the `admin` cluster, which tags it and forwards it
 * to the monitoring cluster
 * @param {Object} kbnServer manager of Kibana services - see `src/server/kbn_server` in Kibana core
 * @param {Object} server HapiJS server instance
 * @param {Object} cloudDetails Cloud details that should be published with Kibana data.
 * @return {Object} the revealed `push` and `flush` modules
 */
export function opsBuffer(kbnServer, server) {
  const config = server.config();
  const interval = config.get('xpack.monitoring.kibana.collection.interval') + 'ms';
  const monitoringTag = config.get('xpack.monitoring.loggingTag');
  const client = server.plugins.elasticsearch.getCluster('admin').createClient({
    plugins: [monitoringBulk]
  });
  const cloudDetector = new CloudDetector();
  // determine the cloud service in the background
  cloudDetector.detectCloudService();

  let lastOp = null;

  return {
    push(event) {
      lastOp = {
        host: event.host,
        rollup: rollupEvent(event, lastOp)
      };

      server.log(['debug', monitoringTag], 'Received Monitoring event data');
    },
    flush() {
      if (!lastOp) { return; }

      // grab the last operation
      const kibana = sourceKibana(kbnServer, config, lastOp.host);
      const rollup = lastOp.rollup;
      const body = [
        // Push the time-based information to .monitoring-kibana-*
        { index: { _type: KIBANA_STATS_TYPE } },
        {
          kibana,
          cloud: cloudDetector.getCloudDetails(),
          ...rollup
        }
      ];

      // reset lastOp
      lastOp = null;

      server.log(['debug', monitoringTag], 'Sending Monitoring payload to Elasticsearch');

      return client.monitoring.bulk({
        system_id: KIBANA_SYSTEM_ID,
        system_api_version: MONITORING_SYSTEM_API_VERSION,
        interval,
        body
      })
      .catch((err) => {
        server.log(['error', monitoringTag], err);
      });
    }
  };
}
