import { events } from './esqueue/constants/events';
import { oncePerServer } from './once_per_server';

function getDocumentPayloadFn(server) {
  const jobQueue = server.plugins.reporting.queue;
  const exportTypesRegistry = server.plugins.reporting.exportTypesRegistry;
  const deprecationLog = message => server.log(['reporting', 'deprecation', 'warning'], message);

  function encodeContent(content, jobType) {
    if (!jobType) {
      return content;
    }

    const exportType = exportTypesRegistry.get(item => item.jobType === jobType);
    switch (exportType.jobContentEncoding) {
      case 'base64':
        return new Buffer(content, 'base64');
      default:
        return content;
    }
  }

  function getCompleted(output, jobType) {
    return {
      statusCode: 200,
      content: encodeContent(output.content, jobType),
      contentType: output.content_type
    };
  }

  function getFailure(output) {
    return {
      statusCode: 500,
      content: {
        message: 'Reporting generation failed',
        reason: output.content
      },
      contentType: 'application/json'
    };
  }

  function getIncomplete(status) {
    return {
      statusCode: 503,
      content: status,
      contentType: 'application/json',
      headers: {
        'retry-after': 30
      }
    };
  }

  function getDocumentPayloadSync(doc, jobType) {
    deprecationLog(`Use of the 'sync' parameter will be removed in the next major version`);

    return new Promise(resolve => {
      // faking sync, wait for the job to be completed
      function completeHandler(completed) {
        // if the completed job matches this job
        if (completed.job.id === doc._id) {
          // remove event listener
          cleanupListeners();
          resolve(getCompleted(completed.output, jobType));
        }
      };

      function errorHandler(failed) {
        if (failed.job.id === doc._id) {
          // remove event listener
          cleanupListeners();
          resolve(getFailure(failed.output));
        }
      };

      function cleanupListeners() {
        jobQueue.removeListener(events.EVENT_WORKER_COMPLETE, completeHandler);
        jobQueue.removeListener(events.EVENT_WORKER_JOB_FAIL, errorHandler);
      }

      jobQueue.on(events.EVENT_WORKER_COMPLETE, completeHandler);
      jobQueue.on(events.EVENT_WORKER_JOB_FAIL, errorHandler);
    });
  }

  return async function getDocumentPayload(doc, opts) {
    const { status, output, jobtype: jobType } = doc._source;

    if (status === 'completed') {
      return getCompleted(output, jobType);
    }

    if (status === 'failed') {
      return getFailure(output);
    }

    if (!opts.sync) {
      // send a 503 indicating that the report isn't completed yet
      return getIncomplete(status);
    }

    return await getDocumentPayloadSync(doc, jobType);
  };
}

export const getDocumentPayloadFactory = oncePerServer(getDocumentPayloadFn);

